% Kinematic car
% Tutorial 1: simulating the car

% Systemtheorie und Regelungstechnik
% Saarland University
% March 2021

clear all; close all; clc;

%% add path to common files
addpath(['..' filesep 'common']);

%% Kinematic car parameters
l = 1; % wheelbase length in m

%% simulation parameters
t_end = 10; % simulation end time
dt = 1/20; % sample time of simulated variables (animation frame rate)

%% simulate

% initial conditions
% state defined as x = [y1, y2, theta]' with respect to input (v,phi);
y10 = 0;
y20 = 0;
theta0 = 0;
x0=[y10;y20;theta0];

% configure solver
% opts = odeset('RelTol',1e-9,'AbsTol',1e-9); % for example
opts = odeset();

% run solver
tt = [0:dt:t_end];
% tt = [0,t_end];
[t,x] = ode45(@(t,x) carODE(t,x,l), tt, x0, opts);

% format simulation output for readability
[y1,y2,theta] = x2state(x'); % see function below
[v,phi] = arrayfun(@carref,t); % apply function ref to each element of array t

%% plot results

% spawn a figure
f=figure;
% set figure size and position
set(f,'Units','normalized','OuterPosition',[0 0 1/3 0.6]);

N=5;i=0;
i=i+1;h(i)=subplot(N,1,i);
p(i)=plot(t,y1);
ylabel('y_1 / m');
i=i+1;h(i)=subplot(N,1,i);
p(i)=plot(t,y2);
ylabel('y_2 / m');
i=i+1;h(i)=subplot(N,1,i);
p(i)=plot(t,theta);
ylabel('\theta / rad');
i=i+1;h(i)=subplot(N,1,i);
p(i)=plot(t,v);
ylabel('v / (m/s)' );
i=i+1;h(i)=subplot(N,1,i);
p(i)=plot(t,phi);
ylabel('\phi / rad');
xlabel('t / s')
set(h,'XGrid','on','YGrid','on');
linkaxes(h,'x');

%% animate results
f1=KinematicCarAnimate(dt, t, l, y1, y2, theta, phi);

%% helper functions

function [v,phi]=carref(t)
% generate reference for control input variables
    v=1;
    phi=0.1*sin(t);
end

function [y1,y2,theta] = x2state(x)
% reformat state vector for readability
    if(size(x,1)== 3)
        y1 = x(1,:);
        y2 = x(2,:);
        theta = x(3,:);
    else
        error('array of column vectors of length 3 expected');
    end    
end

function dxdt = carODE(t,x,l)
% right hand side of dynamic system state equation (1.4)
% See Flatness-based Control: An Introduction, Shaker Verlag, Feb. 2021
% section 1.3, p.4-6

    
    [y1,y2,theta] = x2state(x);

    [v,phi] = carref(t);

    dxdt = [v*cos(theta); ...
            v*sin(theta); ...
            v*tan(phi)/l];
end