classdef KinematicCarDynamicControl < KinematicCarClosedLoopControl
    %KINEMATICCARDYNAMICCONTROL is an dynamic controller for the kinematic car
    %to track the flat output trajectory t\mapsto (y1(t),y2(2)) designed in
    %tutorial 2.
    %See Flatness-based Control: An Introduction, Shaker Verlag, Feb. 2021
    %section 1.7, p.10-14
    
    properties (SetAccess=public, GetAccess=public)
        % controller gains
        K1 double =1;
        K0 double =1;
        lambda;
    end
           
    methods
        function obj = KinematicCarDynamicControl(r,K0,K1)
            %KINEMATICCARCONTROL Construct an instance of this class
                                  
            if nargin>0
                % save reference to reference variable object
                obj.r = r;
                
                obj.K0 = K0;
                obj.K1 = K1;            
            end
            
        end
        
        % TASK 2(i): Implement dynamic controller equations (1.14a) and 
        % (1.14d) to evaluate control variables (v,phi)
        function obj = eval(obj,t,v,theta,y1,y2)
            
            % evaluate reference variables
            eval(obj.r,t);
            
            % dynamic controller
            % (1.14a)     
            obj.lambda = [obj.r.ddy1';obj.r.ddy2']...
                - obj.K1*[v.*cos(theta)-obj.r.dy1';v.*sin(theta)-obj.r.dy2'] ...
                - obj.K0*[y1-obj.r.y1';y2-obj.r.y2'];
            % set control variables
            obj.v = v;
            % (1.14d)
            obj.phi = atan(obj.r.l./v.^2.*dot([-sin(theta); cos(theta)],obj.lambda,1));
        end
        
    end
end

