% Kinematic car
% Tutorial 3: closed-loop dynamic controller

% Systemtheorie und Regelungstechnik
% Saarland University
% March 2021

clear all; close all; clc;

%% add path to common files
addpath(['..' filesep 'common']);
addpath(['..' filesep 't2']); % need ref generator

%% Kinematic car parameters
l = 1; % wheelbase length in m

%% simulation parameters
t_end = 10; % simulation end time
dt = 1/20; % sample time of simulated variables (animation frame rate)

%% simulate

% generate trajectories
t0=0;
t1=t_end;
y10 = 0;
y20 = 0;
dy1r0=1;
dy2r0=0;
theta0 = -pi/6;
v0=sqrt(dy1r0^2+dy2r0^2);

r1=PolyRef(t0,t1,y10,10,dy1r0,0.5,0,0);
r2=PolyRef(t0,t1,y20,5,dy2r0,0,0,0);

% initial conditions
% state defined as x = [y1, y2, theta]' with respect to input (v,phi);
% TASK 1: Consider effect of initial condition errors with open-loop controller
x0=[y10-1;y20-1;theta0;v0];

% Task 2(iii) Initialize the extended ODE solver state appropriately.
x0=[y10-1;y20-1;theta0;v0];

% initialize controllers
r = KinematicCarReference(l,r1,r2); % open-loop controller as reference
% Task 2(iv): Initialize the instance of KinematicCarDynamicControl using the class constructor
c = KinematicCarDynamicControl(r,2,2); % closed-loop controller

% configure solver
opts = odeset('RelTol',1e-3); % default 1e-3
% opts = odeset();

% run solver
tt = [0:dt:t_end];
% tt = [0,t_end];

% TASK 2(v): Modify the call to ode45 to reflect the new carODE.
[t,x] = ode45(@(t,x) carODE(t,x,l,c), tt, x0, opts);

% format simulation output for readability
% TASK 2(vi): evaluate KinematicCarDynamicControl object at desired sample times.
[y1,y2,theta,v] = x2state(x'); % see function below
% evaluate trajectory, reference, and controller at desired sample times
eval(c,t,v,theta,y1,y2);

%% plot results

% spawn a figure
f=figure;
% set figure size and position
set(f,'Units','normalized','OuterPosition',[0 0 1/3 0.6]);

% Task 2(vii): plot controller variables
N=5;i=0;
i=i+1;h(i)=subplot(N,1,i);
p{i}=plot(t,r.y1,t,y1);
ylabel('y_1 / m');
i=i+1;h(i)=subplot(N,1,i);
p{i}=plot(t,r.y2,t,y2);
ylabel('y_2 / m');
i=i+1;h(i)=subplot(N,1,i);
p{i}=plot(t,r.th,t,theta);
ylabel('\theta / rad');
i=i+1;h(i)=subplot(N,1,i);
p{i}=plot(t,r.v,t,c.v);
ylabel('v / (m/s)' );
i=i+1;h(i)=subplot(N,1,i);
p{i}=plot(t,r.phi,t,c.phi);
ylabel('\phi / rad');
xlabel('t / s')
set(h,'XGrid','on','YGrid','on');
linkaxes(h,'x');
colororder(f,{'b','r'})
legend(h(1),'ref','sim')

%% animate results
% TASK 2(viii): pass KinematicCarDynamicControl object for animation
f1=KinematicCarAnimate(dt, t, l, y1, y2, theta, c.phi, c.r.l, c.r.y1, c.r.y2, c.r.th, c.r.phi);

%% helper functions

% TASK 2(ii): Update carODE to accept feedback variables and an argument 
% of type KinematicCarDynamicControl. The integration of the controller’s dynamical
% subsystem described by the differential equation (1.14c) is to be performed by the
% existing ODE solver. Extend the state of the system of ODEs accordingly. You will
% also need to modify helper function x2state

function dxdt = carODE(t,x,l,c)
% right hand side of dynamic system state equation (1.4)
% augmented with controller dynamics (1.14c)
% See Flatness-based Control: An Introduction, Shaker Verlag, Feb. 2021
% section 1.3, p.4-6 and section 1.7 p.10-14
    
    % get current state
    [y1,y2,theta,v] = x2state(x);
    
    % evaluate controller at time t (also evaluates reference variables)
    eval(c,t,v,theta,y1,y2);
    
    % RHS of dynamic controller component (1.14c)
    dvdt = [cos(theta) sin(theta)]*c.lambda;
    
    % extend kinematic car system of ODEs to include \dot{v}
    dxdt = [v*cos(theta); ...
            v*sin(theta); ...
            v*tan(c.phi)/l
            dvdt];
end

function [y1,y2,theta,v] = x2state(x)
% reformat state vector for readability
% extend kinematic car system state to include v
    if(size(x,1)== 4)
        y1 = x(1,:);
        y2 = x(2,:);
        theta = x(3,:);
        v = x(4,:);
    else
        error('array of column vectors of length 4 expected');
    end    
end