% Kinematic car
% Simulation tutorial 5
% Systemtheorie und Regelungstechnik 
clear all; close all; clc;

%% add path to common files
addpath(['..' filesep 'common']);
addpath(['..' filesep 't4']); % need reference from t4

%% Kinematic car parameters
l = 1; % wheelbase length in m

%% simulation parameters
t_end = 10; % simulation end time
dt = 1/20; % sample time of simulated variables (animation frame rate)

%% simulate

% generate trajectories
t0=0;
t1=t_end;

% init controller
yA = [-30;-10]; % initial position
yB = [0;0];    % final position
% yB = [cos(theta0) sin(theta0); -sin(theta0) cos(theta0)]'*[0;2]+yA;

thetaA = 0;% initial orientation
thetaB = 0;  % final orientation
phiA = 0;       % initial steering angle
phiB = 0;       % final steering angle

% initial conditions
% state defined as x = [y1, y2, theta]' with respect to input (v,phi);
y10 = yA(1)-3;
y20 = yA(2)-3;
theta0 = thetaA+0.3;
x0=[y10;y20;theta0];

ref = KinematicCarReference(l,t1,yA,thetaA,phiA,yB,thetaB,phiB);
ctl = KinematicCarQuasistaticControl(1, 1, 2, ref);
% configure solver
opts = odeset('RelTol',1e-3); % default 1e-3
% opts = odeset();

% run solver
tt = [0:dt:t_end];
% tt = [0,t_end];
[t,x] = ode45(@(t,x) carODE(t,x,l,ctl), tt, x0, opts);

% format simulation output for readability
[y1,y2,theta] = x2state(x'); % see function below
% evaluate trajectory and controller at desired sample times
r=arrayfun(@ref.eval,tt);
c=arrayfun(@ctl.eval,tt,y1,y2,theta);

%% plot results

% spawn a figure
f=figure;
% set figure size and position
set(f,'Units','normalized','OuterPosition',[0 0 1/3 0.6]);

N=5;i=0;
i=i+1;h(i)=subplot(N,1,i);
p{i}=plot(t,[r.y1],t,y1);
% p{i}=plot(t,y1);
ylabel('y_1 / m');
i=i+1;h(i)=subplot(N,1,i);
p{i}=plot(t,[r.y2],t,y2);
% p{i}=plot(t,y2);
ylabel('y_2 / m');
i=i+1;h(i)=subplot(N,1,i);
p{i}=plot(t,[r.th],t,theta);
% p{i}=plot(t,theta);
ylabel('\theta / rad');
i=i+1;h(i)=subplot(N,1,i);
p{i}=plot(t,[r.v],t,[c.v]);
ylabel('v / (m/s)' );
i=i+1;h(i)=subplot(N,1,i);
p{i}=plot(t,[r.phi],t,[c.phi]);
ylabel('\phi / rad');
xlabel('t / s')
set(h,'XGrid','on','YGrid','on');
linkaxes(h,'x');
colororder(f,{'b','r'})
legend(h(1),'ref','sim')

%% animate results
f1=KinematicCarAnimate(dt, t, l, y1, y2, theta, [c.phi], ref.l, [r.y1], [r.y2], [r.th], [r.phi]);

%% carODE
function dxdt = carODE(t,x,l,c)
% right hand side of dynamic system state equation (1.4)
% See Flatness-based Control: An Introduction, Shaker Verlag, Feb. 2021
% section 1.3, p.4-6
    
    [y1,y2,theta] = x2state(x);
    
    % evaluate controller
    eval(c,t,y1,y2,theta);
    
    dxdt = [c.v*cos(theta); ...
            c.v*sin(theta); ...
            c.v*tan(c.phi)/l];
end